@@ -749,15 +749,14 @@
 #define GEN5_VERTEXBUFFER_ACCESS_VERTEXDATA     0
 #define GEN5_VERTEXBUFFER_ACCESS_INSTANCEDATA   1
 
-#define GEN5_VFCOMPONENT_NOSTORE      0
-#define GEN5_VFCOMPONENT_STORE_SRC    1
-#define GEN5_VFCOMPONENT_STORE_0      2
-#define GEN5_VFCOMPONENT_STORE_1_FLT  3
-#define GEN5_VFCOMPONENT_STORE_1_INT  4
-#define GEN5_VFCOMPONENT_STORE_VID    5
-#define GEN5_VFCOMPONENT_STORE_IID    6
-#define GEN5_VFCOMPONENT_STORE_PID    7
-
+#define VFCOMPONENT_NOSTORE      0
+#define VFCOMPONENT_STORE_SRC    1
+#define VFCOMPONENT_STORE_0      2
+#define VFCOMPONENT_STORE_1_FLT  3
+#define VFCOMPONENT_STORE_1_INT  4
+#define VFCOMPONENT_STORE_VID    5
+#define VFCOMPONENT_STORE_IID    6
+#define VFCOMPONENT_STORE_PID    7
 
 
 /* Execution Unit (EU) defines
@@ -2759,16 +2758,19 @@ typedef enum {
 
 typedef enum {
 	WM_KERNEL = 0,
-	WM_KERNEL_PROJECTIVE,
+	WM_KERNEL_P,
 
 	WM_KERNEL_MASK,
-	WM_KERNEL_MASK_PROJECTIVE,
+	WM_KERNEL_MASK_P,
 
 	WM_KERNEL_MASKCA,
-	WM_KERNEL_MASKCA_PROJECTIVE,
+	WM_KERNEL_MASKCA_P,
 
-	WM_KERNEL_MASKCA_SRCALPHA,
-	WM_KERNEL_MASKCA_SRCALPHA_PROJECTIVE,
+	WM_KERNEL_MASKSA,
+	WM_KERNEL_MASKSA_P,
+
+	WM_KERNEL_OPACITY,
+	WM_KERNEL_OPACITY_P,
 
 	WM_KERNEL_VIDEO_PLANAR,
 	WM_KERNEL_VIDEO_PACKED,
